@@ -15,6 +15,7 @@
 """Unittests for the manifest_xml.py module."""
 
 import os
+from pathlib import Path
 import platform
 import re
 import tempfile
@@ -51,7 +52,7 @@ INVALID_FS_PATHS = (
     "foo~",
     "blah/foo~",
     # Block Unicode characters that get normalized out by filesystems.
-    "foo\u200Cbar",
+    "foo\u200cbar",
     # Block newlines.
     "f\n/bar",
     "f\r/bar",
@@ -97,36 +98,34 @@ class ManifestParseTestCase(unittest.TestCase):
 
     def setUp(self):
         self.tempdirobj = tempfile.TemporaryDirectory(prefix="repo_tests")
-        self.tempdir = self.tempdirobj.name
-        self.repodir = os.path.join(self.tempdir, ".repo")
-        self.manifest_dir = os.path.join(self.repodir, "manifests")
-        self.manifest_file = os.path.join(
-            self.repodir, manifest_xml.MANIFEST_FILE_NAME
+        self.tempdir = Path(self.tempdirobj.name)
+        self.repodir = self.tempdir / ".repo"
+        self.manifest_dir = self.repodir / "manifests"
+        self.manifest_file = self.repodir / manifest_xml.MANIFEST_FILE_NAME
+        self.local_manifest_dir = (
+            self.repodir / manifest_xml.LOCAL_MANIFESTS_DIR_NAME
         )
-        self.local_manifest_dir = os.path.join(
-            self.repodir, manifest_xml.LOCAL_MANIFESTS_DIR_NAME
-        )
-        os.mkdir(self.repodir)
-        os.mkdir(self.manifest_dir)
+        self.repodir.mkdir()
+        self.manifest_dir.mkdir()
 
         # The manifest parsing really wants a git repo currently.
-        gitdir = os.path.join(self.repodir, "manifests.git")
-        os.mkdir(gitdir)
-        with open(os.path.join(gitdir, "config"), "w") as fp:
-            fp.write(
-                """[remote "origin"]
+        gitdir = self.repodir / "manifests.git"
+        gitdir.mkdir()
+        (gitdir / "config").write_text(
+            """[remote "origin"]
         url = https://localhost:0/manifest
 """
-            )
+        )
 
     def tearDown(self):
         self.tempdirobj.cleanup()
 
     def getXmlManifest(self, data):
         """Helper to initialize a manifest for testing."""
-        with open(self.manifest_file, "w", encoding="utf-8") as fp:
-            fp.write(data)
-        return manifest_xml.XmlManifest(self.repodir, self.manifest_file)
+        self.manifest_file.write_text(data, encoding="utf-8")
+        return manifest_xml.XmlManifest(
+            str(self.repodir), str(self.manifest_file)
+        )
 
     @staticmethod
     def encodeXmlAttr(attr):
@@ -243,12 +242,14 @@ class XmlManifestTests(ManifestParseTestCase):
 
     def test_link(self):
         """Verify Link handling with new names."""
-        manifest = manifest_xml.XmlManifest(self.repodir, self.manifest_file)
-        with open(os.path.join(self.manifest_dir, "foo.xml"), "w") as fp:
-            fp.write("<manifest></manifest>")
+        manifest = manifest_xml.XmlManifest(
+            str(self.repodir), str(self.manifest_file)
+        )
+        (self.manifest_dir / "foo.xml").write_text("<manifest></manifest>")
         manifest.Link("foo.xml")
-        with open(self.manifest_file) as fp:
-            self.assertIn('<include name="foo.xml" />', fp.read())
+        self.assertIn(
+            '<include name="foo.xml" />', self.manifest_file.read_text()
+        )
 
     def test_toxml_empty(self):
         """Verify the ToXml() helper."""
@@ -406,10 +407,9 @@ class IncludeElementTests(ManifestParseTestCase):
 
     def test_revision_default(self):
         """Check handling of revision attribute."""
-        root_m = os.path.join(self.manifest_dir, "root.xml")
-        with open(root_m, "w") as fp:
-            fp.write(
-                """
+        root_m = self.manifest_dir / "root.xml"
+        root_m.write_text(
+            """
 <manifest>
   <remote name="test-remote" fetch="http://localhost" />
   <default remote="test-remote" revision="refs/heads/main" />
@@ -418,17 +418,34 @@ class IncludeElementTests(ManifestParseTestCase):
   <project name="root-name2" path="root-path2" />
 </manifest>
 """
-            )
-        with open(os.path.join(self.manifest_dir, "stable.xml"), "w") as fp:
-            fp.write(
-                """
+        )
+        (self.manifest_dir / "stable.xml").write_text(
+            """
 <manifest>
+  <include name="man1.xml" />
+  <include name="man2.xml" revision="stable-branch2" />
   <project name="stable-name1" path="stable-path1" />
   <project name="stable-name2" path="stable-path2" revision="stable-branch2" />
 </manifest>
 """
-            )
-        include_m = manifest_xml.XmlManifest(self.repodir, root_m)
+        )
+        (self.manifest_dir / "man1.xml").write_text(
+            """
+<manifest>
+  <project name="man1-name1" />
+  <project name="man1-name2" revision="stable-branch3" />
+</manifest>
+"""
+        )
+        (self.manifest_dir / "man2.xml").write_text(
+            """
+<manifest>
+  <project name="man2-name1" />
+  <project name="man2-name2" revision="stable-branch3" />
+</manifest>
+"""
+        )
+        include_m = manifest_xml.XmlManifest(str(self.repodir), str(root_m))
         for proj in include_m.projects:
             if proj.name == "root-name1":
                 # Check include revision not set on root level proj.
@@ -442,12 +459,19 @@ class IncludeElementTests(ManifestParseTestCase):
             if proj.name == "stable-name2":
                 # Check stable proj revision can override include node.
                 self.assertEqual("stable-branch2", proj.revisionExpr)
+            if proj.name == "man1-name1":
+                self.assertEqual("stable-branch", proj.revisionExpr)
+            if proj.name == "man1-name2":
+                self.assertEqual("stable-branch3", proj.revisionExpr)
+            if proj.name == "man2-name1":
+                self.assertEqual("stable-branch2", proj.revisionExpr)
+            if proj.name == "man2-name2":
+                self.assertEqual("stable-branch3", proj.revisionExpr)
 
     def test_group_levels(self):
-        root_m = os.path.join(self.manifest_dir, "root.xml")
-        with open(root_m, "w") as fp:
-            fp.write(
-                """
+        root_m = self.manifest_dir / "root.xml"
+        root_m.write_text(
+            """
 <manifest>
   <remote name="test-remote" fetch="http://localhost" />
   <default remote="test-remote" revision="refs/heads/main" />
@@ -456,25 +480,23 @@ class IncludeElementTests(ManifestParseTestCase):
   <project name="root-name2" path="root-path2" groups="r2g1,r2g2" />
 </manifest>
 """
-            )
-        with open(os.path.join(self.manifest_dir, "level1.xml"), "w") as fp:
-            fp.write(
-                """
+        )
+        (self.manifest_dir / "level1.xml").write_text(
+            """
 <manifest>
   <include name="level2.xml" groups="level2-group" />
   <project name="level1-name1" path="level1-path1" />
 </manifest>
 """
-            )
-        with open(os.path.join(self.manifest_dir, "level2.xml"), "w") as fp:
-            fp.write(
-                """
+        )
+        (self.manifest_dir / "level2.xml").write_text(
+            """
 <manifest>
   <project name="level2-name1" path="level2-path1" groups="l2g1,l2g2" />
 </manifest>
 """
-            )
-        include_m = manifest_xml.XmlManifest(self.repodir, root_m)
+        )
+        include_m = manifest_xml.XmlManifest(str(self.repodir), str(root_m))
         for proj in include_m.projects:
             if proj.name == "root-name1":
                 # Check include group not set on root level proj.
@@ -492,6 +514,41 @@ class IncludeElementTests(ManifestParseTestCase):
                 # Check level2 proj group not removed.
                 self.assertIn("l2g1", proj.groups)
 
+    def test_group_levels_with_extend_project(self):
+        root_m = self.manifest_dir / "root.xml"
+        root_m.write_text(
+            """
+<manifest>
+  <remote name="test-remote" fetch="http://localhost" />
+  <default remote="test-remote" revision="refs/heads/main" />
+  <include name="man1.xml" groups="top-group1" />
+  <include name="man2.xml" groups="top-group2" />
+</manifest>
+"""
+        )
+        (self.manifest_dir / "man1.xml").write_text(
+            """
+<manifest>
+  <project name="project1" path="project1" />
+</manifest>
+"""
+        )
+        (self.manifest_dir / "man2.xml").write_text(
+            """
+<manifest>
+  <extend-project name="project1" groups="eg1" />
+</manifest>
+"""
+        )
+        include_m = manifest_xml.XmlManifest(str(self.repodir), str(root_m))
+        proj = include_m.projects[0]
+        # Check project has inherited group via project element.
+        self.assertIn("top-group1", proj.groups)
+        # Check project has inherited group via extend-project element.
+        self.assertIn("top-group2", proj.groups)
+        # Check project has set group via extend-project element.
+        self.assertIn("eg1", proj.groups)
+
     def test_allow_bad_name_from_user(self):
         """Check handling of bad name attribute from the user's input."""
 
@@ -510,9 +567,8 @@ class IncludeElementTests(ManifestParseTestCase):
             manifest.ToXml()
 
         # Setup target of the include.
-        target = os.path.join(self.tempdir, "target.xml")
-        with open(target, "w") as fp:
-            fp.write("<manifest></manifest>")
+        target = self.tempdir / "target.xml"
+        target.write_text("<manifest></manifest>")
 
         # Include with absolute path.
         parse(os.path.abspath(target))
@@ -526,12 +582,9 @@ class IncludeElementTests(ManifestParseTestCase):
         def parse(name):
             name = self.encodeXmlAttr(name)
             # Setup target of the include.
-            with open(
-                os.path.join(self.manifest_dir, "target.xml"),
-                "w",
-                encoding="utf-8",
-            ) as fp:
-                fp.write(f'<manifest><include name="{name}"/></manifest>')
+            (self.manifest_dir / "target.xml").write_text(
+                f'<manifest><include name="{name}"/></manifest>'
+            )
 
             manifest = self.getXmlManifest(
                 """
@@ -578,18 +631,18 @@ class ProjectElementTests(ManifestParseTestCase):
             manifest.projects[0].name: manifest.projects[0].groups,
             manifest.projects[1].name: manifest.projects[1].groups,
         }
-        self.assertCountEqual(
-            result["test-name"], ["name:test-name", "all", "path:test-path"]
+        self.assertEqual(
+            result["test-name"], {"name:test-name", "all", "path:test-path"}
         )
-        self.assertCountEqual(
+        self.assertEqual(
             result["extras"],
-            ["g1", "g2", "g1", "name:extras", "all", "path:path"],
+            {"g1", "g2", "name:extras", "all", "path:path"},
         )
         groupstr = "default,platform-" + platform.system().lower()
-        self.assertEqual(groupstr, manifest.GetGroupsStr())
+        self.assertEqual(groupstr, manifest.GetManifestGroupsStr())
         groupstr = "g1,g2,g1"
         manifest.manifestProject.config.SetString("manifest.groups", groupstr)
-        self.assertEqual(groupstr, manifest.GetGroupsStr())
+        self.assertEqual(groupstr, manifest.GetManifestGroupsStr())
 
     def test_set_revision_id(self):
         """Check setting of project's revisionId."""
@@ -1214,6 +1267,166 @@ class ExtendProjectElementTests(ManifestParseTestCase):
         self.assertEqual(len(manifest.projects), 1)
         self.assertEqual(manifest.projects[0].upstream, "bar")
 
+    def test_extend_project_copyfiles(self):
+        manifest = self.getXmlManifest(
+            """
+<manifest>
+  <remote name="default-remote" fetch="http://localhost" />
+  <default remote="default-remote" revision="refs/heads/main" />
+  <project name="myproject" />
+  <extend-project name="myproject">
+    <copyfile src="foo" dest="bar" />
+  </extend-project>
+</manifest>
+"""
+        )
+        self.assertEqual(list(manifest.projects[0].copyfiles)[0].src, "foo")
+        self.assertEqual(list(manifest.projects[0].copyfiles)[0].dest, "bar")
+        self.assertEqual(
+            sort_attributes(manifest.ToXml().toxml()),
+            '<?xml version="1.0" ?><manifest>'
+            '<remote fetch="http://localhost" name="default-remote"/>'
+            '<default remote="default-remote" revision="refs/heads/main"/>'
+            '<project name="myproject">'
+            '<copyfile dest="bar" src="foo"/>'
+            "</project>"
+            "</manifest>",
+        )
+
+    def test_extend_project_duplicate_copyfiles(self):
+        root_m = self.manifest_dir / "root.xml"
+        root_m.write_text(
+            """
+<manifest>
+  <remote name="test-remote" fetch="http://localhost" />
+  <default remote="test-remote" revision="refs/heads/main" />
+  <project name="myproject" />
+  <include name="man1.xml" />
+  <include name="man2.xml" />
+</manifest>
+"""
+        )
+        (self.manifest_dir / "man1.xml").write_text(
+            """
+<manifest>
+  <include name="common.xml" />
+</manifest>
+"""
+        )
+        (self.manifest_dir / "man2.xml").write_text(
+            """
+<manifest>
+  <include name="common.xml" />
+</manifest>
+"""
+        )
+        (self.manifest_dir / "common.xml").write_text(
+            """
+<manifest>
+  <extend-project name="myproject">
+    <copyfile dest="bar" src="foo"/>
+  </extend-project>
+</manifest>
+"""
+        )
+        manifest = manifest_xml.XmlManifest(str(self.repodir), str(root_m))
+        self.assertEqual(len(manifest.projects[0].copyfiles), 1)
+        self.assertEqual(list(manifest.projects[0].copyfiles)[0].src, "foo")
+        self.assertEqual(list(manifest.projects[0].copyfiles)[0].dest, "bar")
+
+    def test_extend_project_linkfiles(self):
+        manifest = self.getXmlManifest(
+            """
+<manifest>
+  <remote name="default-remote" fetch="http://localhost" />
+  <default remote="default-remote" revision="refs/heads/main" />
+  <project name="myproject" />
+  <extend-project name="myproject">
+    <linkfile src="foo" dest="bar" />
+  </extend-project>
+</manifest>
+"""
+        )
+        self.assertEqual(list(manifest.projects[0].linkfiles)[0].src, "foo")
+        self.assertEqual(list(manifest.projects[0].linkfiles)[0].dest, "bar")
+        self.assertEqual(
+            sort_attributes(manifest.ToXml().toxml()),
+            '<?xml version="1.0" ?><manifest>'
+            '<remote fetch="http://localhost" name="default-remote"/>'
+            '<default remote="default-remote" revision="refs/heads/main"/>'
+            '<project name="myproject">'
+            '<linkfile dest="bar" src="foo"/>'
+            "</project>"
+            "</manifest>",
+        )
+
+    def test_extend_project_duplicate_linkfiles(self):
+        root_m = self.manifest_dir / "root.xml"
+        root_m.write_text(
+            """
+<manifest>
+  <remote name="test-remote" fetch="http://localhost" />
+  <default remote="test-remote" revision="refs/heads/main" />
+  <project name="myproject" />
+  <include name="man1.xml" />
+  <include name="man2.xml" />
+</manifest>
+"""
+        )
+        (self.manifest_dir / "man1.xml").write_text(
+            """
+<manifest>
+  <include name="common.xml" />
+</manifest>
+"""
+        )
+        (self.manifest_dir / "man2.xml").write_text(
+            """
+<manifest>
+  <include name="common.xml" />
+</manifest>
+"""
+        )
+        (self.manifest_dir / "common.xml").write_text(
+            """
+<manifest>
+  <extend-project name="myproject">
+    <linkfile dest="bar" src="foo"/>
+  </extend-project>
+</manifest>
+"""
+        )
+        manifest = manifest_xml.XmlManifest(str(self.repodir), str(root_m))
+        self.assertEqual(len(manifest.projects[0].linkfiles), 1)
+        self.assertEqual(list(manifest.projects[0].linkfiles)[0].src, "foo")
+        self.assertEqual(list(manifest.projects[0].linkfiles)[0].dest, "bar")
+
+    def test_extend_project_annotations(self):
+        manifest = self.getXmlManifest(
+            """
+<manifest>
+  <remote name="default-remote" fetch="http://localhost" />
+  <default remote="default-remote" revision="refs/heads/main" />
+  <project name="myproject" />
+  <extend-project name="myproject">
+    <annotation name="foo" value="bar" />
+  </extend-project>
+</manifest>
+"""
+        )
+        self.assertEqual(manifest.projects[0].annotations[0].name, "foo")
+        self.assertEqual(manifest.projects[0].annotations[0].value, "bar")
+        self.assertEqual(
+            sort_attributes(manifest.ToXml().toxml()),
+            '<?xml version="1.0" ?><manifest>'
+            '<remote fetch="http://localhost" name="default-remote"/>'
+            '<default remote="default-remote" revision="refs/heads/main"/>'
+            '<project name="myproject">'
+            '<annotation name="foo" value="bar"/>'
+            "</project>"
+            "</manifest>",
+        )
+
 
 class NormalizeUrlTests(ManifestParseTestCase):
     """Tests for normalize_url() in manifest_xml.py"""
